/* $Id: compare.c 643 2016-08-03 22:40:45Z bhockney $ */

/* (C) 2004-2016 by Bob Hockney <zeus@ix.netcom.com>    *
 * Based on fwlogwatch written by                       *
 * Boris Wesslowski <Wesslowski@CERT.Uni-Stuttgart.DE>  *
 *                                                      *
 * sorting functions for wfwl_syslog                    *
 *                                                      *
 * This code is distributed under the terms of GNU GPL  */

#include <string.h>

#include <main.h>
#include <compare.h>
#include <output.h>
#include <utils.h>

extern struct conn_data *first;
extern struct options opt;
extern struct field_order *fields;

/* compares two rows in data and indicates if they need to be sorted */
unsigned char compare(struct conn_data *op1, struct conn_data *op2)
{
  unsigned char cond = 0;
  int comp = 0;
  unsigned long n_op1 = 0;
  unsigned long n_op2 = 0;
  double d_op1 = 0;
  double d_op2 = 0;
  char * s_op1 = NULL;
  char * s_op2 = NULL;
  unsigned short int f_op1 = AF_INET;
  unsigned short int f_op2 = AF_INET;
  struct in_addr * addr_op1 = NULL;
  struct in_addr * addr_op2 = NULL;
#if HAVE_INET_NTOP
  struct in6_addr * addr6_op1 = NULL;
  struct in6_addr * addr6_op2 = NULL;
#endif

  switch(opt.sortfield) {
  /* begin numeric data */
  case SORT_LOCAL_TIME:
  case SORT_START_TIME:
  case SORT_END_TIME:
  case SORT_COUNT:
  case SORT_OOB_PROTOCOL:
  case SORT_OOB_FAMILY:
  case SORT_FWMARK:
  case SORT_PROTOCOL:
  case SORT_IP_TOTLEN:
  case SORT_IP_ID:
  case SORT_IP_IHL:
  case SORT_IP_TOS:
  case SORT_IP_TTL:
  case SORT_IP6_PAY_LEN:
  case SORT_IP6_FLOWLABEL:
  case SORT_CSUM:
  case SORT_SOURCEPORT:
  case SORT_DESTPORT:
  case SORT_TCP_SEQ:
  case SORT_TCP_ACKSEQ:
  case SORT_TCP_WINDOW:
  case SORT_TCP_URGP:
  case SORT_TCP_OPTS:
  case SORT_UDP_LEN:
  case SORT_FRAGOFF:
  case SORT_ICMP_TYPE:
  case SORT_ICMP_CODE:
  case SORT_ICMP_ECHOID:
  case SORT_ICMP_ECHOSEQ:
  case SORT_ICMP_MTU:
  case SORT_AHESP_SPI:
  case SORT_METRO_CODE_SRC:
  case SORT_ACCURACY_RADIUS_SRC:
  case SORT_IS_ANONYMOUS_PROXY_SRC:
  case SORT_IS_SATELLITE_PROVIDER_SRC:
  case SORT_METRO_CODE_DST:
  case SORT_ACCURACY_RADIUS_DST:
  case SORT_IS_ANONYMOUS_PROXY_DST:
  case SORT_IS_SATELLITE_PROVIDER_DST:
    switch (opt.sortfield) {
      case SORT_LOCAL_TIME:
        n_op1 = op1->local_time; n_op2 = op2->local_time;
        break;
      case SORT_START_TIME:
        n_op1 = (unsigned long)op1->start_time; n_op2 = (unsigned long)op2->start_time;
        break;
      case SORT_END_TIME:
        n_op1 = (unsigned long)op1->end_time; n_op2 = (unsigned long)op2->end_time;
        break;
      case SORT_COUNT:
        n_op1 = op1->count; n_op2 = op2->count;
        break;
      case SORT_OOB_PROTOCOL:
        n_op1 = op1->oob_protocol; n_op2 = op2->oob_protocol;
        break;
      case SORT_OOB_FAMILY:
        n_op1 = op1->family; n_op2 = op2->family;
        break;
      case SORT_FWMARK:
        n_op1 = op1->fwmark; n_op2 = op2->fwmark;
        break;
      case SORT_PROTOCOL:
        n_op1 = op1->protocol; n_op2 = op2->protocol;
        break;
      case SORT_IP_TOTLEN:
        n_op1 = op1->totlen; n_op2 = op2->totlen;
        break;
      case SORT_IP_ID:
        n_op1 = op1->ipid; n_op2 = op2->ipid;
        break;
      case SORT_IP_IHL:
        n_op1 = op1->ihl; n_op2 = op2->ihl;
        break;
      case SORT_IP_TOS:
        n_op1 = op1->tos; n_op2 = op2->tos;
        break;
      case SORT_IP_TTL:
        n_op1 = op1->ttl; n_op2 = op2->ttl;
        break;
      case SORT_IP6_PAY_LEN:
        n_op1 = op1->pay_len; n_op2 = op2->pay_len;
        break;
      case SORT_IP6_FLOWLABEL:
        n_op1 = op1->flowlabel; n_op2 = op2->flowlabel;
        break;
      case SORT_CSUM:
        n_op1 = op1->csum; n_op2 = op2->csum;
        break;
      case SORT_SOURCEPORT:
        n_op1 = op1->sport; n_op2 = op2->sport;
        break;
      case SORT_DESTPORT:
        n_op1 = op1->dport; n_op2 = op2->dport;
        break;
      case SORT_TCP_SEQ:
        n_op1 = op1->tcp_seq; n_op2 = op2->tcp_seq;
        break;
      case SORT_TCP_ACKSEQ:
        n_op1 = op1->tcp_ack_seq; n_op2 = op2->tcp_ack_seq;
        break;
      case SORT_TCP_WINDOW:
        n_op1 = op1->tcp_window; n_op2 = op2->tcp_window;
        break;
      case SORT_TCP_URGP:
        n_op1 = op1->tcp_urgp; n_op2 = op2->tcp_urgp;
        break;
      case SORT_TCP_OPTS:
        n_op1 = op1->flags; n_op2 = op2->flags;
        break;
      case SORT_FRAGOFF:
        n_op1 = op1->fragoff; n_op2 = op2->fragoff;
        break;
      case SORT_UDP_LEN:
        n_op1 = op1->udp_len; n_op2 = op2->udp_len;
        break;
      case SORT_ICMP_TYPE:
        n_op1 = op1->icmp_type; n_op2 = op2->icmp_type;
        break;
      case SORT_ICMP_CODE:
        n_op1 = op1->icmp_code; n_op2 = op2->icmp_code;
        break;
      case SORT_ICMP_ECHOID:
        n_op1 = op1->icmp_echoid; n_op2 = op2->icmp_echoid;
        break;
      case SORT_ICMP_ECHOSEQ:
        n_op1 = op1->icmp_echoseq; n_op2 = op2->icmp_echoseq;
        break;
      case SORT_ICMP_MTU:
        n_op1 = op1->icmp_mtu; n_op2 = op2->icmp_mtu;
        break;
      case SORT_AHESP_SPI:
        n_op1 = op1->ahesp_spi; n_op2 = op2->ahesp_spi;
        break;
      case SORT_METRO_CODE_SRC:
        n_op1 = op1->geoip_data_src->metro_code; n_op2 = op2->geoip_data_src->metro_code;
        break;
      case SORT_ACCURACY_RADIUS_SRC:
        n_op1 = op1->geoip_data_src->accuracy_radius; n_op2 = op2->geoip_data_src->accuracy_radius;
        break;
      case SORT_IS_ANONYMOUS_PROXY_SRC:
        n_op1 = op1->geoip_data_src->is_anonymous_proxy; n_op2 = op2->geoip_data_src->is_anonymous_proxy;
        break;
      case SORT_IS_SATELLITE_PROVIDER_SRC:
        n_op1 = op1->geoip_data_src->is_satellite_provider; n_op2 = op2->geoip_data_src->is_satellite_provider;
        break;
      case SORT_METRO_CODE_DST:
        n_op1 = op1->geoip_data_dst->metro_code; n_op2 = op2->geoip_data_dst->metro_code;
        break;
      case SORT_ACCURACY_RADIUS_DST:
        n_op1 = op1->geoip_data_dst->accuracy_radius; n_op2 = op2->geoip_data_dst->accuracy_radius;
        break;
      case SORT_IS_ANONYMOUS_PROXY_DST:
        n_op1 = op1->geoip_data_dst->is_anonymous_proxy; n_op2 = op2->geoip_data_dst->is_anonymous_proxy;
        break;
      case SORT_IS_SATELLITE_PROVIDER_DST:
        n_op1 = op1->geoip_data_dst->is_satellite_provider; n_op2 = op2->geoip_data_dst->is_satellite_provider;
        break;
    }

    if (opt.sortmode == ORDER_ASCENDING) {
      if (n_op1 > n_op2) cond++;
    } else {
      if (n_op1 < n_op2) cond++;
    }
    break;
  /* end numeric data */
  /* begin double data */
  case SORT_LATITUDE_SRC:
  case SORT_LONGITUDE_SRC:
  case SORT_LATITUDE_DST:
  case SORT_LONGITUDE_DST:
    switch(opt.sortfield) {
      case SORT_LATITUDE_SRC:
        d_op1 = op1->geoip_data_src->latitude; d_op2 = op2->geoip_data_src->latitude;
        break;
      case SORT_LONGITUDE_SRC:
        d_op1 = op1->geoip_data_src->longitude; d_op2 = op2->geoip_data_src->longitude;
        break;
      case SORT_LATITUDE_DST:
        d_op1 = op1->geoip_data_dst->latitude; d_op2 = op2->geoip_data_dst->latitude;
        break;
      case SORT_LONGITUDE_DST:
        d_op1 = op1->geoip_data_dst->longitude; d_op2 = op2->geoip_data_dst->longitude;
        break;
    }

    if (opt.sortmode == ORDER_ASCENDING) {
      if (d_op1 > d_op2) cond++;
    } else {
      if (d_op1 < d_op2) cond++;
    }
  break;
  /* end double data */
  /* begin sort id */
  case SORT_ID:
    comp = strncmp(op1->filename, op2->filename, SHORTLEN);
    if (opt.sortmode == ORDER_ASCENDING) {
      if (comp > 0 || (comp == 0 && op1->linenum > op2->linenum)) cond++;
    } else {
      if (comp < 0 || (comp == 0 && op1->linenum < op2->linenum)) cond++;
    }
    break;
  /* begin string data */
  case SORT_PREFIX: /* PREFIX vs CHAINLABEL */
  case SORT_SNAME:
  case SORT_LOCAL_HOST:
  case SORT_IN_IF:
  case SORT_OUT_IF:
  case SORT_MAC_SADDR:
  case SORT_MAC_DADDR:
  case SORT_MAC:
  case SORT_CONTINENT_CODE_SRC:
  case SORT_CONTINENT_NAME_SRC:
  case SORT_COUNTRY_ISO_CODE_SRC:
  case SORT_COUNTRY_NAME_SRC:
  case SORT_SUBDIVISION_1_ISO_CODE_SRC:
  case SORT_SUBDIVISION_1_NAME_SRC:
  case SORT_SUBDIVISION_2_ISO_CODE_SRC:
  case SORT_SUBDIVISION_2_NAME_SRC:
  case SORT_CITY_NAME_SRC:
  case SORT_TIME_ZONE_SRC:
  case SORT_REGISTERED_COUNTRY_GEONAME_ID_SRC:
  case SORT_REPRESENTED_COUNTRY_GEONAME_ID_SRC:
  case SORT_POSTAL_CODE_SRC:
  case SORT_CONTINENT_CODE_DST:
  case SORT_CONTINENT_NAME_DST:
  case SORT_COUNTRY_ISO_CODE_DST:
  case SORT_COUNTRY_NAME_DST:
  case SORT_SUBDIVISION_1_ISO_CODE_DST:
  case SORT_SUBDIVISION_1_NAME_DST:
  case SORT_SUBDIVISION_2_ISO_CODE_DST:
  case SORT_SUBDIVISION_2_NAME_DST:
  case SORT_CITY_NAME_DST:
  case SORT_TIME_ZONE_DST:
  case SORT_REGISTERED_COUNTRY_GEONAME_ID_DST:
  case SORT_REPRESENTED_COUNTRY_GEONAME_ID_DST:
  case SORT_POSTAL_CODE_DST:
    switch(opt.sortfield) {
      case SORT_PREFIX: /* PREFIX vs CHAINLABEL */
        s_op1 = op1->log_label; s_op2 = op2->log_label;
        break;
      case SORT_SNAME:
        s_op1 = op1->filename; s_op2 = op2->filename;
        break;
      case SORT_LOCAL_HOST:
        s_op1 = op1->hostname; s_op2 = op2->hostname;
        break;
      case SORT_IN_IF:
        s_op1 = op1->inif; s_op2 = op2->inif;
        break;
      case SORT_OUT_IF:
        s_op1 = op1->outif; s_op2 = op2->outif;
        break;
      case SORT_MAC_SADDR:
        s_op1 = op1->mac_saddr_str; s_op2 = op2->mac_saddr_str;
        break;
      case SORT_MAC_DADDR:
        s_op1 = op1->mac_daddr_str; s_op2 = op2->mac_daddr_str;
        break;
      case SORT_MAC:
        s_op1 = op1->raw_mac;; s_op2 = op2->raw_mac;;
        break;
      case SORT_CONTINENT_CODE_SRC:
        s_op1 = op1->geoip_data_src->continent_code;; s_op2 = op2->geoip_data_src->continent_code;;
        break;
      case SORT_CONTINENT_NAME_SRC:
        s_op1 = op1->geoip_data_src->continent_name;; s_op2 = op2->geoip_data_src->continent_name;;
        break;
      case SORT_COUNTRY_ISO_CODE_SRC:
        s_op1 = op1->geoip_data_src->country_iso_code;; s_op2 = op2->geoip_data_src->country_iso_code;;
        break;
      case SORT_COUNTRY_NAME_SRC:
        s_op1 = op1->geoip_data_src->country_name;; s_op2 = op2->geoip_data_src->country_name;;
        break;
      case SORT_SUBDIVISION_1_ISO_CODE_SRC:
        s_op1 = op1->geoip_data_src->subdivision_1_iso_code;; s_op2 = op2->geoip_data_src->subdivision_1_iso_code;;
        break;
      case SORT_SUBDIVISION_1_NAME_SRC:
        s_op1 = op1->geoip_data_src->subdivision_1_name;; s_op2 = op2->geoip_data_src->subdivision_1_name;;
        break;
      case SORT_SUBDIVISION_2_ISO_CODE_SRC:
        s_op1 = op1->geoip_data_src->subdivision_2_iso_code;; s_op2 = op2->geoip_data_src->subdivision_2_iso_code;;
        break;
      case SORT_SUBDIVISION_2_NAME_SRC:
        s_op1 = op1->geoip_data_src->subdivision_2_name;; s_op2 = op2->geoip_data_src->subdivision_2_name;;
        break;
      case SORT_CITY_NAME_SRC:
        s_op1 = op1->geoip_data_src->city_name;; s_op2 = op2->geoip_data_src->city_name;;
        break;
      case SORT_TIME_ZONE_SRC:
        s_op1 = op1->geoip_data_src->time_zone;; s_op2 = op2->geoip_data_src->time_zone;;
        break;
      case SORT_REGISTERED_COUNTRY_GEONAME_ID_SRC:
        s_op1 = op1->geoip_data_src->registered_country_geoname_id;; s_op2 = op2->geoip_data_src->registered_country_geoname_id;;
        break;
      case SORT_REPRESENTED_COUNTRY_GEONAME_ID_SRC:
        s_op1 = op1->geoip_data_src->represented_country_geoname_id;; s_op2 = op2->geoip_data_src->represented_country_geoname_id;;
        break;
      case SORT_POSTAL_CODE_SRC:
        s_op1 = op1->geoip_data_src->postal_code;; s_op2 = op2->geoip_data_src->postal_code;;
        break;
      case SORT_CONTINENT_CODE_DST:
        s_op1 = op1->geoip_data_dst->continent_code;; s_op2 = op2->geoip_data_dst->continent_code;;
        break;
      case SORT_CONTINENT_NAME_DST:
        s_op1 = op1->geoip_data_dst->continent_name;; s_op2 = op2->geoip_data_dst->continent_name;;
        break;
      case SORT_COUNTRY_ISO_CODE_DST:
        s_op1 = op1->geoip_data_dst->country_iso_code;; s_op2 = op2->geoip_data_dst->country_iso_code;;
        break;
      case SORT_COUNTRY_NAME_DST:
        s_op1 = op1->geoip_data_dst->country_name;; s_op2 = op2->geoip_data_dst->country_name;;
        break;
      case SORT_SUBDIVISION_1_ISO_CODE_DST:
        s_op1 = op1->geoip_data_dst->subdivision_1_iso_code;; s_op2 = op2->geoip_data_dst->subdivision_1_iso_code;;
        break;
      case SORT_SUBDIVISION_1_NAME_DST:
        s_op1 = op1->geoip_data_dst->subdivision_1_name;; s_op2 = op2->geoip_data_dst->subdivision_1_name;;
        break;
      case SORT_SUBDIVISION_2_ISO_CODE_DST:
        s_op1 = op1->geoip_data_dst->subdivision_2_iso_code;; s_op2 = op2->geoip_data_dst->subdivision_2_iso_code;;
        break;
      case SORT_SUBDIVISION_2_NAME_DST:
        s_op1 = op1->geoip_data_dst->subdivision_2_name;; s_op2 = op2->geoip_data_dst->subdivision_2_name;;
        break;
      case SORT_CITY_NAME_DST:
        s_op1 = op1->geoip_data_dst->city_name;; s_op2 = op2->geoip_data_dst->city_name;;
        break;
      case SORT_TIME_ZONE_DST:
        s_op1 = op1->geoip_data_dst->time_zone;; s_op2 = op2->geoip_data_dst->time_zone;;
        break;
      case SORT_REGISTERED_COUNTRY_GEONAME_ID_DST:
        s_op1 = op1->geoip_data_dst->registered_country_geoname_id;; s_op2 = op2->geoip_data_dst->registered_country_geoname_id;;
        break;
      case SORT_REPRESENTED_COUNTRY_GEONAME_ID_DST:
        s_op1 = op1->geoip_data_dst->represented_country_geoname_id;; s_op2 = op2->geoip_data_dst->represented_country_geoname_id;;
        break;
      case SORT_POSTAL_CODE_DST:
        s_op1 = op1->geoip_data_dst->postal_code;; s_op2 = op2->geoip_data_dst->postal_code;;
        break;
    }

    if (opt.sortmode == ORDER_ASCENDING) {
      if (strncmp(s_op1, s_op2, SHORTLEN) > 0) cond++;
    } else {
      if (strncmp(s_op1, s_op2, SHORTLEN) < 0) cond++;
    }
    break;
  /* end string data */
  /* begin ip data */
  case SORT_SOURCEHOST:
  case SORT_DESTHOST:
  case SORT_ICMP_GW:
    switch (opt.sortfield) {
      case SORT_SOURCEHOST:
        f_op1 = op1->family;
        f_op2 = op2->family;
        addr_op1 = &op1->shost;
        addr_op2 = &op2->shost;
#if HAVE_INET_NTOP
        addr6_op1 = &op1->shost6;
        addr6_op2 = &op2->shost6;
#endif
        break;
      case SORT_DESTHOST:
        f_op1 = op1->family;
        f_op2 = op2->family;
        addr_op1 = &op1->dhost;
        addr_op2 = &op2->dhost;
#if HAVE_INET_NTOP
        addr6_op1 = &op1->dhost6;
        addr6_op2 = &op2->dhost6;
#endif
        break;
      case SORT_ICMP_GW:
        f_op1 = op1->family;
        f_op2 = op2->family;
        addr_op1 = &op1->icmp_gw;
        addr_op2 = &op2->icmp_gw;
        break;
    }

    if (opt.sortmode == ORDER_ASCENDING) {
      if (f_op1 != f_op2) {
        if (f_op1 > f_op2) cond++;
      } else
        if (f_op1 == AF_INET && (ntohl(addr_op1->s_addr) > ntohl(addr_op2->s_addr))) cond++;
#if HAVE_INET_NTOP
        else if (f_op1 == AF_INET6 && memcmp(addr6_op1->s6_addr, addr6_op2->s6_addr,sizeof(struct in6_addr)) >0) cond++;
#endif
    } else {
      if (f_op1 != f_op2) {
        if (f_op1 < f_op2) cond++;
      } else
        if (f_op1 == AF_INET && (ntohl(addr_op1->s_addr) < ntohl(addr_op2->s_addr))) cond++;
#if HAVE_INET_NTOP
        else if (f_op1 == AF_INET6 && memcmp(addr6_op1->s6_addr, addr6_op2->s6_addr,sizeof(struct in6_addr)) <0) cond++;
#endif
    }
    break;
  /* end ip data */

  default:
    if (opt.verbose >= VERBOSE_INFO)
      fprintf(stderr, "Unrecognized sort key %i\n", opt.sortfield);
  }

  return cond;
}

/* sorts parsed data */
struct conn_data *wfwl_sort(struct conn_data *list) {
  struct conn_data *p, *q, *e, *tail;
  int size, merges, psize, qsize, i;

  if(list != NULL) {
    size = 1;
    while(1) {
      p = list;
      list = tail = NULL;
      merges = 0;
      while (p != NULL) {
        merges++;
        q = p;
        psize = 0;
        for (i = 0; i < size; i++) {
          psize++;
          q = q->next;
          if (q == NULL) break;
        }
        qsize = size;
        while (psize > 0 || ((qsize > 0) && (q != NULL))) {
          if (psize == 0) {
            e = q; q = q->next; qsize--;
          } else if (qsize == 0 || (q == NULL)) {
            e = p; p = p->next; psize--;
        /* compare function looks at struct opt to determine how to sort */
          } else if (compare(p,q) <= 0) {
            e = p; p = p->next; psize--;
          } else {
            e = q; q = q->next; qsize--;
          }
          if (tail != NULL) {
            tail->next = e;
          } else {
            list = e;
          }
          tail = e;
        } /* end "while (psize > 0 || ((qsize > 0) && (q != NULL)))" */
        p = q;
      } /* end "while (p != NULL)" */
      tail->next = NULL;
      if (merges <= 1)
        return list;
      size *= 2;
    } /* end "while(1)" */
  } else { /* end "while(list != NULL)" */
    return NULL;
  }
}
