/* $Id: netfilter.l 643 2016-08-03 22:40:45Z bhockney $ */

/* (C) 2004-2016 by Bob Hockney <zeus@ix.netcom.com>    *
 * Based on fwlogwatch written by                       *
 * Boris Wesslowski <Wesslowski@CERT.Uni-Stuttgart.DE>  *
 *                                                      *
 * netfilter log parser for wfwl_syslog                 *
 *                                                      *
 * This code is distributed under the terms of GNU GPL  */

%option prefix="nf"
%option outfile="lex.yy.c"
%option noyywrap
%option nounput
%option noinput

%{
#include <unistd.h>
#include <string.h>
#include <ctype.h>

#include <main.h>
#include <utils.h>
#include <getdate.h>

#ifndef IP_DF
#define IP_DF 0x4000
#endif
#ifndef IP_MF
#define IP_MF 0x2000
#endif
#ifndef AF_INET6
#define AF_INET6 0
#endif

extern struct options opt;

void nf_parse_start_hn(char *input);
void nf_parse_start_nh(char *input);
void nf_parse_isostart_hn(char *input);
void nf_parse_rfc5424start_hn(char *input);
void nf_parse_prefix(char *input, unsigned char mode);
void nf_parse_ip(char *input, unsigned char mode, unsigned short int ip_type);
void nf_parse_proto(char *input);
%}

MONTH	"Jan"|"Feb"|"Mar"|"Apr"|"May"|"Jun"|"Jul"|"Aug"|"Sep"|"Oct"|"Nov"|"Dec"
STRING	[a-zA-Z][a-zA-Z0-9.-]*
SPACE	[ ]
PREFIX	[ -~-]+
LOGHOST	[0-9.a-zA-Z()_:-]+
DIGIT	[0-9]
NUMBER	{DIGIT}+
KERNEL	"kernel:"|"klogd:"|"[kernel]"|"ulogd["{NUMBER}"]:"
TIME	"["{SPACE}*{DIGIT}+"."{DIGIT}+"]"
OCTET	{DIGIT}{1,3}
PORT	{DIGIT}{1,5}
HEXDIGIT	[0-9a-fA-F]
HEXNUM	"0x"?{HEXDIGIT}+
MAC	({HEXDIGIT}{2}":"){5}{HEXDIGIT}{2}
PROTO	"TCP"|"UDP"|"UDPLITE"|"ICMP"|"ESP"|"AH"|"ICMPv6"|{NUMBER}

%s have_kernel have_if have_proto

%%

{MONTH}[ ]{1,2}{DIGIT}{1,2}[ ]{DIGIT}{2}:{DIGIT}{2}:{DIGIT}{2}[ ]{KERNEL}	nf_parse_start_nh(nftext); BEGIN(have_kernel);
{MONTH}[ ]{1,2}{DIGIT}{1,2}[ ]{DIGIT}{2}:{DIGIT}{2}:{DIGIT}{2}[ ]{LOGHOST}	nf_parse_start_hn(nftext);
{MONTH}[ ]{1,2}{DIGIT}{1,2}[ ]{DIGIT}{2}:{DIGIT}{2}:{DIGIT}{2}	nf_parse_start_nh(nftext);
{DIGIT}{4}[ ]{MONTH}[ ]{1,2}{DIGIT}{1,2}[ ]{DIGIT}{2}:{DIGIT}{2}:{DIGIT}{2}[ ]{LOGHOST}	nf_parse_start_hn(nftext+5);
{DIGIT}{4}[ ]{MONTH}[ ]{1,2}{DIGIT}{1,2}[ ]{DIGIT}{2}:{DIGIT}{2}:{DIGIT}{2}	nf_parse_start_nh(nftext+5);
{DIGIT}{4}:{DIGIT}{2}:{DIGIT}{2}-{DIGIT}{2}:{DIGIT}{2}:{DIGIT}{2}[ ]{LOGHOST}	nf_parse_isostart_hn(nftext);
{DIGIT}{4}"-"{DIGIT}{2}"-"{DIGIT}{2}"T"{DIGIT}{2}":"{DIGIT}{2}":"{DIGIT}{2}("."{DIGIT}{1,6})?((("+"|"-"){DIGIT}{2}":"{DIGIT}{2})|"Z")[ ]{LOGHOST} { 
	nf_parse_rfc5424start_hn(nftext);}
" "{KERNEL}" "{TIME}			BEGIN(have_kernel);
" "{KERNEL}				BEGIN(have_kernel);
<have_kernel>" "{PREFIX}"IN="{STRING}?	nf_parse_prefix(nftext+1, NF_OPT_PREFIX); BEGIN(have_if);
"IN="{STRING}?		nf_parse_prefix(nftext+3, NF_OPT_NOPREFIX); BEGIN(have_if);
"OUT="{STRING}?		xstrncpy(opt.line->outif, nftext+4, SHORTLEN);
"PHYSIN="{STRING}?	/* ignore */
"PHYSOUT="{STRING}?	/* ignore */
"MAC="(({HEXDIGIT}{HEXDIGIT}:){13,33}{HEXDIGIT}{HEXDIGIT})?	xstrncpy(opt.line->raw_mac, nftext+4, SHORTLEN);
"MACSRC="{MAC}		xstrncpy(opt.line->mac_saddr_str, nftext+7, SHORTLEN);
"MACDST="{MAC}		xstrncpy(opt.line->mac_daddr_str, nftext+7, SHORTLEN);
"MACPROTO="{HEXDIGIT}+	opt.line->oob_protocol = strtoul(nftext+9, NULL, 16);
"SRC="{OCTET}"."{OCTET}"."{OCTET}"."{OCTET}	nf_parse_ip(nftext+4, NF_OPT_SRC, AF_INET);
"DST="{OCTET}"."{OCTET}"."{OCTET}"."{OCTET}	nf_parse_ip(nftext+4, NF_OPT_DST, AF_INET);
"SRC="({HEXDIGIT}{0,4}:){2,7}{HEXDIGIT}{0,4}	nf_parse_ip(nftext+4, NF_OPT_SRC, AF_INET6);
"DST="({HEXDIGIT}{0,4}:){2,7}{HEXDIGIT}{0,4}	nf_parse_ip(nftext+4, NF_OPT_DST, AF_INET6);
<have_if>"LEN="{NUMBER}	{
        		if (opt.line->family == AF_INET) {
        		    opt.line->totlen = strtoul(nftext+4, NULL, 0);
                        } else {
        		    opt.line->pay_len = strtoul(nftext+4, NULL, 0);
        		}
        		}
"TOS="({HEXNUM}|"00")	opt.line->tos |= strtol(nftext+4, NULL, 0);
"PREC="{HEXNUM}		opt.line->tos |= strtol(nftext+5, NULL, 0);
"TTL="{NUMBER}		opt.line->ttl = atoi(nftext+4);
<have_if>"ID="{NUMBER}	opt.line->ipid = strtoul(nftext+3, NULL, 0);
<have_if>"ID:"{HEXDIGIT}{8}	opt.line->ipid = strtoul(nftext+3, NULL, 16);
"CE"			/* ignore - should never appear; CE is part of tos, not fragoff */
"DF"			opt.line->fragoff |= IP_DF;
"MF"			opt.line->fragoff |= IP_MF;
"FRAG:"{SPACE}?{NUMBER}		opt.line->fragoff |= atoi(nftext+5);
"PROTO="{PROTO}		{ nf_parse_proto(nftext+6); BEGIN(have_proto); }
"INCOMPLETE ["{NUMBER}" bytes]"	/* ignore */
"TYPE="{NUMBER}		{ opt.line->icmp_type = atoi(nftext+5); opt.parser=opt.parser|NF_TYPE; }
"CODE="{NUMBER}		opt.line->icmp_code = atoi(nftext+5);
"SEQ="{NUMBER}		{
        		if (opt.line->protocol == 6) {
        		    opt.line->tcp_seq = strtoul(nftext+4, NULL, 0);
                        } else {
        		    opt.line->icmp_echoseq = strtoul(nftext+4, NULL, 0);
        		}
        		}
"ACK="{NUMBER}		opt.line->tcp_ack_seq = strtoul(nftext+4, NULL, 0);
"SPT="{PORT}		{ opt.line->sport = atol(nftext+4); opt.parser=opt.parser|NF_SPT; }
"DPT="{PORT}		{ opt.line->dport = atol(nftext+4); opt.parser=opt.parser|NF_DPT; }
"WINDOW="{NUMBER}	opt.line->tcp_window = atol(nftext+7);
"RES="{HEXNUM}		/* ignore */
"URGP="{NUMBER}		opt.line->tcp_urgp = atol(nftext+5);
"URG"			opt.line->flags |= TCP_URG;
"ACK"			opt.line->flags |= TCP_ACK;
"PSH"			opt.line->flags |= TCP_PSH;
"RST"			opt.line->flags |= TCP_RST;
"SYN"			opt.line->flags |= TCP_SYN;
"FIN"			opt.line->flags |= TCP_FIN;
"ECE"			/* ignore */
"CWR"			/* ignore */
"OPT ("{HEXDIGIT}+")"	/* ignore */
"OPT (TRUNCATED)"	/* ignore */
"SPI="{HEXNUM}		opt.line->ahesp_spi = strtoul(nftext+4, NULL, 0);
"GATEWAY="{OCTET}"."{OCTET}"."{OCTET}"."{OCTET}	convert_ip(AF_INET, nftext+8, &opt.line->icmp_gw, NULL, NULL);
"MTU="{NUMBER}		opt.line->icmp_mtu = atoi(nftext+4);
"TC="{NUMBER}		opt.line->tos |= strtol(nftext+3, NULL, 0);
"HOPLIMIT="{NUMBER}	opt.line->ttl = atoi(nftext+9);
"FLOWLBL="{NUMBER}	opt.line->flowlabel = strtoul(nftext+8, NULL, 0);
"LEN="{NUMBER}		opt.line->udp_len = atol(nftext+4);
"ID="{NUMBER}		opt.line->icmp_echoid = atoi(nftext+3);
"PARAMETER="{NUMBER}	/* ignore */
"POINTER="{HEXDIGIT}{8}	/* ignore */
"MARK="{HEXNUM}		opt.line->fwmark = strtoul(nftext+5, NULL, 16);
"UID="			/* ignore */
"GID="			/* ignore */
"TRUNCATED"		/* ignore */
"[".+"]"		/* ignore */
{SPACE}			/* ignore whitespace */
[\n]			return 0;
{STRING}		if(opt.verbose >= VERBOSE_ERROR) fprintf(stderr, "Unrecognized token: %s\n", nftext);
.			if(opt.verbose >= VERBOSE_ERROR) fprintf(stderr, "Unrecognized character: %s\n", nftext);

%%

void nf_parse_start_hn(char *input)
{
  int retval, day, hour, minute, second;
  char smonth[4];

  retval = sscanf(input,
        	  "%3s %2d %2d:%2d:%2d %" SHOSTLEN_S "s",
        	  smonth, &day, &hour, &minute, &second,
        	  opt.line->hostname);
  if (retval != 6) {
    return;
  }

  build_time(smonth, day, hour, minute, second);

  opt.parser=opt.parser|NF_DATE;
}
 
void nf_parse_start_nh(char *input)
{
  int retval, day, hour, minute, second;
  char smonth[4];

  retval = sscanf(input,
		  "%3s %2d %2d:%2d:%2d",
		  smonth, &day, &hour, &minute, &second);
  if (retval != 5) {
    return;
  }

  build_time(smonth, day, hour, minute, second);
  xstrncpy(opt.line->hostname, "-", SHORTLEN);

  opt.parser=opt.parser|NF_DATE;
}

void nf_parse_isostart_hn(char *input)
{
  int retval, year, month, day, hour, minute, second;
  struct tm *t;

  retval = sscanf(input,
		  "%4d:%2d:%2d-%2d:%2d:%2d %" SHOSTLEN_S "s",
		  &year, &month, &day, &hour, &minute, &second,
		  opt.line->hostname);
  if (retval != 7) {
    return;
  }

  t = xmalloc(sizeof(struct tm));
  t->tm_year = year-1900;
  t->tm_mon = month-1;
  t->tm_mday = day;
  t->tm_hour = hour;
  t->tm_min = minute;
  t->tm_sec = second;
  t->tm_isdst = -1;
  opt.line->time = mktime(t);
  free(t);

  opt.parser=opt.parser|NF_DATE;
}

void nf_parse_rfc5424start_hn(char *input)
{
  int retval, year, month, day, hour, minute, second;
  struct tm *t;

  retval = sscanf(input, "%4d-%2d-%2dT%2d:%2d:%2d%*s %32s", &year, &month, &day, &hour, &minute, &second, opt.line->hostname);
  if (retval != 7) {
    return;
  }

  t = xmalloc(sizeof(struct tm));
  t->tm_year = year - 1900;
  t->tm_mon = month - 1;
  t->tm_mday = day;
  t->tm_hour = hour;
  t->tm_min = minute;
  t->tm_sec = second;
  t->tm_isdst = -1;
  opt.line->time = mktime(t);
  free(t);

  opt.parser = opt.parser | NF_DATE;
}


void nf_parse_prefix(char *input, unsigned char mode)
{
  size_t i=0;

  if(mode == NF_OPT_PREFIX) {
    i=strlen(input);
    while (i > 0) {
      if(input[i] == '=') {
	if (input[i-1] == 'N') {
	  if (input[i-2] == 'I') {
	    input[i-3] = '\0'; /* don't include trailing space in log_label */
            xstrncpy(opt.line->log_label, input, SHORTLEN);
            xstrncpy(opt.line->interface, input+i+1, SHORTLEN); /* drop this line */
            xstrncpy(opt.line->inif, input+i+1, SHORTLEN);
            break;
          }
        }
      }
      i--;
    }
  } else {
    xstrncpy(opt.line->log_label, "-", SHORTLEN);
    xstrncpy(opt.line->interface, input, SHORTLEN); /* drop this line */
    xstrncpy(opt.line->inif, input, SHORTLEN);
  }

  opt.parser=opt.parser|NF_IN;
}

void nf_parse_ip(char *input, unsigned char mode, unsigned short int ip_type)
{
#ifdef HAVE_INET_NTOP
  void *shost6 = &opt.line->shost6;
  void *dhost6 = &opt.line->dhost6;
#else
  void *shost6 = NULL;
  void *dhost6 = NULL;
#endif
  if (mode == NF_OPT_SRC) {
    if(convert_ip(ip_type, input, &opt.line->shost, shost6, opt.line->shost_s) == IN_ADDR_ERROR) return;
    opt.line->family = ip_type;
    opt.parser=opt.parser|NF_SRC;
  } else {
    if(convert_ip(ip_type, input, &opt.line->dhost, dhost6, opt.line->dhost_s) == IN_ADDR_ERROR) return;
    opt.parser=opt.parser|NF_DST;
  }
}

void nf_parse_proto(char *input)
{
  if(isdigit((int)input[0])) {
    opt.line->protocol = atoi(input);
  } else {
    if(strncmp(input, "TCP", 3) == 0) opt.line->protocol = 6;
    else if(strncmp(input, "UDP", 3) == 0) opt.line->protocol = 17;
    else if(strncmp(input, "ICMPv6", 6) == 0) opt.line->protocol = 58;
    else if(strncmp(input, "ICMP", 4) == 0) opt.line->protocol = 1;
    else if(strncmp(input, "ESP", 3) == 0) opt.line->protocol = 50;
    else if(strncmp(input, "AH", 2) == 0) opt.line->protocol = 51;
  }

  if (opt.line->protocol != 0)
    opt.parser=opt.parser|NF_PROTO;
}

unsigned char flex_netfilter(char *input, int linenum)
{
  struct yy_buffer_state *state;
  
  opt.parser = 0;
  init_line();
  state = nf_scan_string(input);
  nflex();
  nf_delete_buffer(state);

  opt.line->count = 1;

  if (((opt.line->protocol == 6) || (opt.line->protocol == 17)) && (opt.parser == (NF_DATE|NF_PROTO|NF_IN|NF_SRC|NF_DST|NF_SPT|NF_DPT))) {
    return PARSE_OK;
  }
  if (((opt.line->protocol == 1) || (opt.line->protocol == 58)) && (opt.parser == (NF_DATE|NF_PROTO|NF_IN|NF_SRC|NF_DST|NF_TYPE))) {
    return PARSE_OK;
  }
  if (!((opt.line->protocol == 6) || (opt.line->protocol == 17) ||
        (opt.line->protocol == 1) || (opt.line->protocol == 58)) && 
        ((opt.parser & (NF_DATE|NF_PROTO|NF_IN|NF_SRC|NF_DST)) ==
           (NF_DATE|NF_PROTO|NF_IN|NF_SRC|NF_DST))) {
    return PARSE_OK;
  }
  if(opt.verbose >= VERBOSE_ERROR)
    fprintf(stderr, "netfilter parse error in line %d, ignoring.\n", linenum);
  if(opt.verbose >= VERBOSE_DEBUG)
    fprintf(stderr, "input was: \"%s\"\n", input);
  return PARSE_WRONG_FORMAT;
}
