/* $Id: output.c 643 2016-08-03 22:40:45Z bhockney $ */

/* (C) 2004-2016 by Bob Hockney <zeus@ix.netcom.com>    *
 * Based on fwlogwatch written by                       *
 * Boris Wesslowski <Wesslowski@CERT.Uni-Stuttgart.DE>  *
 *                                                      *
 * output function for wfwl_syslog                      *
 *                                                      *
 * This code is distributed under the terms of GNU GPL  */

#include <string.h>

#include <main.h>
#include <output.h>
#include <utils.h>
#include <database.h>

extern struct options opt;
extern struct conn_data *first;
extern struct field_order *fields;
extern struct select_sum select_sum;

/* prints difference between first and last logged packets for line */
void output_timediff(time_t start, time_t end, char *td)
{
  time_t diff;
  int part;
  char tmp[4];

  diff = end - start;
  if (diff <= 0) {
    snprintf(td, 2, "-");
    return;
  }

  part = diff / 86400; /* days */
  snprintf(td, TIMESIZE, "%02d:", part);
  diff = diff % 86400;

  part = diff / 3600; /* hours */
  snprintf(tmp, 4, "%02d:", part);
  strncat(td, tmp, 4);
  diff = diff % 3600;

  part = diff / 60; /* minutes */
  snprintf(tmp, 4, "%02d:", part);
  strncat(td, tmp, 4);

  part = diff % 60; /* seconds */
  snprintf(tmp, 3, "%02d", part);
  strncat(td, tmp, 3);
}

void output_tcp_opts(struct conn_data *input, char *buf)
{
  if (input->protocol == 6 ) { 
    if ((input->flags & (TCP_ACK|TCP_FIN|TCP_RST|TCP_PSH|TCP_URG)) != 0) {
      if (input->flags & TCP_SYN) { strcpy(buf, "s"); } else { strcpy (buf, "-"); }
      if (input->flags & TCP_ACK) { strcat(buf, "a"); } else { strcat (buf, "-"); }
      if (input->flags & TCP_FIN) { strcat(buf, "f"); } else { strcat (buf, "-"); }
      if (input->flags & TCP_RST) { strcat(buf, "r"); } else { strcat (buf, "-"); }
      if (input->flags & TCP_PSH) { strcat(buf, "p"); } else { strcat (buf, "-"); }
      if (input->flags & TCP_URG) { strcat(buf, "u"); } else { strcat (buf, "-"); }
    } else {
      if (input->flags & TCP_SYN) {
        strcpy(buf, "SYN");
      } else {
        strcpy(buf, "-");
      }
    }
  } else {
    strcpy(buf, "");
  }
}

/* output one report line, tab delimited */
void output_wfwl(struct conn_data *input, FILE *fd)
{
  char *proto, buf[SHORTLEN];
  char hostname[SHORTLEN], service[SHORTLEN];
  struct field_order *flds;

  flds = fields;

  while (flds != NULL) {
      switch (flds->field) {
      case COL_ID:
        fprintf(fd, "%lu", input->linenum);
        break;
      case COL_SNAME:
        fprintf(fd, "%s", input->filename);
        break;
      case COL_COUNT:
        fprintf(fd, "%lu", input->count);
        break;
      case COL_IP_PROTO:
        proto = resolve_protocol(input->protocol);
        fprintf(fd, "%s", proto);
        break;
      case COL_IN_IF:
        fprintf(fd, "%s", input->inif);
        break;
      case COL_OUT_IF:
        fprintf(fd, "%s", input->outif);
        break;
      case COL_TCP_SEQ:
        if (input->protocol == 6)
        fprintf(fd, "%u", input->tcp_seq);
        break;
      case COL_TCP_ACKSEQ:
        if (input->protocol == 6)
        fprintf(fd, "%u", input->tcp_ack_seq);
        break;
      case COL_ICMP_TYPE:
        if (input->protocol == 1 || input->protocol == 58)
          fprintf(fd, "%i", input->icmp_type);
        break;
      case COL_ICMP_CODE:
        if (input->protocol == 1 || input->protocol == 58)
          fprintf(fd, "%i", input->icmp_code);
        break;
      case COL_FWMARK:
        fprintf(fd, "%u", input->fwmark);
        break;
      case COL_MAC_SADDR:
        fprintf(fd, "%s", input->mac_saddr_str);
        break;
      case COL_MAC_DADDR:
        fprintf(fd, "%s", input->mac_daddr_str);
        break;
      case COL_MAC:
        fprintf(fd, "%s", input->raw_mac);
        break;
      case COL_START_TIME:
        /* Earliest */
        fprintf(fd, "%li", (long int)input->start_time);
        break;
      case COL_END_TIME:
        /* Latest */
        fprintf(fd, "%li", (long int)input->end_time);
        break;
      case COL_LOCAL_HOST:
        fprintf(fd, "%s", input->hostname);
        break;
      case COL_IP_LEN:
        fprintf(fd, "%u", input->totlen);
        break;
      case COL_IP_IHL:
        fprintf(fd, "%i", input->ihl);
        break;
      case COL_OOB_FAMILY:
        fprintf(fd, "%u", input->family);
        break;
      case COL_OOB_PROTOCOL:
        fprintf(fd, "%u", input->oob_protocol);
        break;
      case COL_SOURCEHOST:
        fprintf(fd, "%s", input->shost_s);
        break;
      case COL_DESTHOST:
        fprintf(fd, "%s", input->dhost_s);
        break;
      case COL_SOURCEHOST_NAME:
        if (strlen(input->shostname)) {
          fputs((char *)&input->shostname, fd);
        } else {
          get_hostname(input->shost_s, hostname);
          fputs(hostname, fd);
        }
        break;
      case COL_DESTHOST_NAME:
        if (strlen(input->dhostname)) {
          fputs((char *)&input->dhostname, fd);
        } else {
          get_hostname(input->dhost_s, hostname);
          fputs(hostname, fd);
        }
        break;
      case COL_SOURCE_SERVICE:
        if (strlen(input->sservice)) {
          fputs((char *)&input->sservice, fd);
        } else {
          get_service_name(input->protocol, input->sport, service);
          fputs(service, fd);
        }
        break;
      case COL_DEST_SERVICE:
        if (strlen(input->dservice)) {
          fputs((char *)&input->dservice, fd);
        } else {
          get_service_name(input->protocol, input->dport, service);
          fputs(service, fd);
        }
        break;
      case COL_SOURCEPORT:
        if (input->protocol == 6 || input->protocol == 17)
          fprintf(fd, "%u", input->sport);
        break;
      case COL_DESTPORT:
        if (input->protocol == 6 || input->protocol == 17)
          fprintf(fd, "%u", input->dport);
        break;
      case COL_LOG_PREFIX:
        fprintf(fd, "%s", input->log_label);
        break;
      case COL_IP_TOS:
        fprintf(fd, "%i", input->tos);
        break;
      case COL_IP_TTL:
        fprintf(fd, "%i", input->ttl);
        break;
      case COL_IP_CSUM:
        fprintf(fd, "%i", input->csum);
        break;
      case COL_IP_ID:
        fprintf(fd, "%u", input->ipid);
        break;
      case COL_IP_FRAGOFF:
        fprintf(fd, "%i", input->fragoff);
        break;
      case COL_IP6_PAY_LEN:
        fprintf(fd, "%u", input->pay_len);
        break;
      case COL_IP6_FLOWLABEL:
        fprintf(fd, "%u", input->flowlabel);
        break;
      case COL_TCP_WINDOW:
        if (input->protocol == 6)
        fprintf(fd, "%u", input->tcp_window);
        break;
      case COL_TCP_URGP:
        if (input->protocol == 6)
        fprintf(fd, "%i", input->tcp_urgp);
        break;
      case COL_UDP_LEN:
        if (input->protocol == 17)
        fprintf(fd, "%u", input->udp_len);
        break;
      case COL_ICMP_ECHOID:
        if ((input->protocol == 1 && (input->icmp_type == 8 || input->icmp_type == 0 )) ||
            (input->protocol == 58 && (input->icmp_type == 128 || input->icmp_type == 129 )))
        fprintf(fd, "%i", input->icmp_echoid);
        break;
      case COL_ICMP_ECHOSEQ:
        if ((input->protocol == 1 && (input->icmp_type == 8 || input->icmp_type == 0 )) ||
            (input->protocol == 58 && (input->icmp_type == 128 || input->icmp_type == 129 )))
        fprintf(fd, "%i", input->icmp_echoseq);
        break;
      case COL_ICMP_GATEWAY:
        if (input->protocol == 1 && input->icmp_type == 5)
        fprintf(fd, "%s", inet_ntoa(input->icmp_gw));
        break;
      case COL_ICMP_MTU:
        if (input->protocol == 1 && input->icmp_type == 3 && input->icmp_code == 4 )
        fprintf(fd, "%i", input->icmp_mtu);
        break;
      case COL_AHESP_SPI:
        if (input->protocol == 50 || input->protocol == 51 )
        fprintf(fd, "%u", input->ahesp_spi);
        break;
      case COL_LOCAL_TIME:
        fprintf(fd, "%li", (long int)input->local_time);
        break;
      case COL_TCP_OPTIONS:
        output_tcp_opts(input, buf);
        fprintf(fd, "%s", buf);
        break;
      case COL_CONTINENT_CODE_SRC:
        fprintf(fd, "%s", input->geoip_data_src->continent_code);
        break;
      case COL_CONTINENT_NAME_SRC:
        fprintf(fd, "%s", input->geoip_data_src->continent_name);
        break;
      case COL_COUNTRY_ISO_CODE_SRC:
        fprintf(fd, "%s", input->geoip_data_src->country_iso_code);
        break;
      case COL_COUNTRY_NAME_SRC:
        fprintf(fd, "%s", input->geoip_data_src->country_name);
        break;
      case COL_SUBDIVISION_1_ISO_CODE_SRC:
        fprintf(fd, "%s", input->geoip_data_src->subdivision_1_iso_code);
        break;
      case COL_SUBDIVISION_1_NAME_SRC:
        fprintf(fd, "%s", input->geoip_data_src->subdivision_1_name);
        break;
      case COL_SUBDIVISION_2_ISO_CODE_SRC:
        fprintf(fd, "%s", input->geoip_data_src->subdivision_2_iso_code);
        break;
      case COL_SUBDIVISION_2_NAME_SRC:
        fprintf(fd, "%s", input->geoip_data_src->subdivision_2_name);
        break;
      case COL_CITY_NAME_SRC:
        fprintf(fd, "%s", input->geoip_data_src->city_name);
        break;
      case COL_METRO_CODE_SRC:
        fprintf(fd, "%u", input->geoip_data_src->metro_code);
        break;
      case COL_ACCURACY_RADIUS_SRC:
        fprintf(fd, "%u", input->geoip_data_src->accuracy_radius);
        break;
      case COL_TIME_ZONE_SRC:
        fprintf(fd, "%s", input->geoip_data_src->time_zone);
        break;
      case COL_REGISTERED_COUNTRY_GEONAME_ID_SRC:
        fprintf(fd, "%s", input->geoip_data_src->registered_country_geoname_id);
        break;
      case COL_REPRESENTED_COUNTRY_GEONAME_ID_SRC:
        fprintf(fd, "%s", input->geoip_data_src->represented_country_geoname_id);
        break;
      case COL_IS_ANONYMOUS_PROXY_SRC:
        fprintf(fd, "%u", input->geoip_data_src->is_anonymous_proxy);
        break;
      case COL_IS_SATELLITE_PROVIDER_SRC:
        fprintf(fd, "%u", input->geoip_data_src->is_satellite_provider);
        break;
      case COL_POSTAL_CODE_SRC:
        fprintf(fd, "%s", input->geoip_data_src->postal_code);
        break;
      case COL_LATITUDE_SRC:
        fprintf(fd, "%.7g", input->geoip_data_src->latitude);
        break;
      case COL_LONGITUDE_SRC:
        fprintf(fd, "%.7g", input->geoip_data_src->longitude);
        break;
      case COL_CONTINENT_CODE_DST:
        fprintf(fd, "%s", input->geoip_data_dst->continent_code);
        break;
      case COL_CONTINENT_NAME_DST:
        fprintf(fd, "%s", input->geoip_data_dst->continent_name);
        break;
      case COL_COUNTRY_ISO_CODE_DST:
        fprintf(fd, "%s", input->geoip_data_dst->country_iso_code);
        break;
      case COL_COUNTRY_NAME_DST:
        fprintf(fd, "%s", input->geoip_data_dst->country_name);
        break;
      case COL_SUBDIVISION_1_ISO_CODE_DST:
        fprintf(fd, "%s", input->geoip_data_dst->subdivision_1_iso_code);
        break;
      case COL_SUBDIVISION_1_NAME_DST:
        fprintf(fd, "%s", input->geoip_data_dst->subdivision_1_name);
        break;
      case COL_SUBDIVISION_2_ISO_CODE_DST:
        fprintf(fd, "%s", input->geoip_data_dst->subdivision_2_iso_code);
        break;
      case COL_SUBDIVISION_2_NAME_DST:
        fprintf(fd, "%s", input->geoip_data_dst->subdivision_2_name);
        break;
      case COL_CITY_NAME_DST:
        fprintf(fd, "%s", input->geoip_data_dst->city_name);
        break;
      case COL_METRO_CODE_DST:
        fprintf(fd, "%u", input->geoip_data_dst->metro_code);
        break;
      case COL_ACCURACY_RADIUS_DST:
        fprintf(fd, "%u", input->geoip_data_dst->accuracy_radius);
        break;
      case COL_TIME_ZONE_DST:
        fprintf(fd, "%s", input->geoip_data_dst->time_zone);
        break;
      case COL_REGISTERED_COUNTRY_GEONAME_ID_DST:
        fprintf(fd, "%s", input->geoip_data_dst->registered_country_geoname_id);
        break;
      case COL_REPRESENTED_COUNTRY_GEONAME_ID_DST:
        fprintf(fd, "%s", input->geoip_data_dst->represented_country_geoname_id);
        break;
      case COL_IS_ANONYMOUS_PROXY_DST:
        fprintf(fd, "%u", input->geoip_data_dst->is_anonymous_proxy);
        break;
      case COL_IS_SATELLITE_PROVIDER_DST:
        fprintf(fd, "%u", input->geoip_data_dst->is_satellite_provider);
        break;
      case COL_POSTAL_CODE_DST:
        fprintf(fd, "%s", input->geoip_data_dst->postal_code);
        break;
      case COL_LATITUDE_DST:
        fprintf(fd, "%.7g", input->geoip_data_dst->latitude);
        break;
      case COL_LONGITUDE_DST:
        fprintf(fd, "%.7g", input->geoip_data_dst->longitude);
        break;
      default:
        break;
    } /* end switch */

    if (flds->next != NULL) {
      fprintf(fd, "\t");
    }

    flds = flds->next;
  } /* end while (flds != NULL) */

  fprintf(fd, "\n");

}

/* determines output mode and calls function to output report */
void show_list(FILE *fd)
{
  struct conn_data *this;
  unsigned int max = 0, skip;

  this = first;

  for (skip = opt.begin; this != NULL && skip > 0; skip-- ) {
    this = this->next;
  }

  while ((this != NULL) && (opt.max == 0 || max < opt.max)) {

    output_wfwl(this, fd);

    if (opt.max != 0)
      max++;

    this = this->next;

  }
}
